# This file is part of daf_relation.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations

__all__ = (
    "ColumnContainer",
    "ColumnRangeLiteral",
    "ColumnExpressionSequence",
    "ColumnInContainer",
)

import dataclasses
from abc import ABC, abstractmethod
from collections.abc import Sequence, Set
from typing import TYPE_CHECKING

from lsst.utils.classes import cached_getter

from ._predicate import Predicate
from ._tag import ColumnTag

if TYPE_CHECKING:
    from .._engine import Engine
    from ._expression import ColumnExpression


class ColumnContainer(ABC):
    """A abstract base class and factory for expressions that represent
    containers of multiple column values.

    `ColumnContainer` inheritance is closed to the types already provided by
    this package.  These concrete types can all be constructed via factory
    methods on `ColumnContainer` itself, so the derived types themselves only
    need to be referenced when writing `match` expressions that process an
    expression tree.  See :ref:`lsst.daf.relation-overview-extensibility` for
    rationale and details.
    """

    def __init_subclass__(cls) -> None:
        assert cls.__name__ in {
            "ColumnRangeLiteral",
            "ColumnExpressionSequence",
        }, "ColumnContainer inheritance is closed to predefined types in daf_relation."

    dtype: type | None
    """The Python type of the elements in the container (`type` or `None`).

    Interpretation of this attribute is up to the `Engine` or other algorithms
    that operate on the expression tree; it is ignored by all code in the
    `lsst.daf.relation` package.
    """

    @property
    @abstractmethod
    def columns_required(self) -> Set[ColumnTag]:
        """Columns required by this expression
        (`~collections.abc.Set` [ `ColumnTag` ]).

        This includes columns required by expressions nested within this one.
        """
        raise NotImplementedError()

    @abstractmethod
    def is_supported_by(self, engine: Engine) -> bool:
        """Test whether the given engine is capable of evaluating this
        expression.

        Parameters
        ----------
        engine : `Engine`
            Engine to test.

        Returns
        -------
        supported : `bool`
            Whether the engine supports this expression and all expressions
            nested within it.
        """
        raise NotImplementedError()

    def contains(self, item: ColumnExpression) -> ColumnInContainer:
        """Construct a boolean column expression that tests whether a scalar
        expression is present in this container expression.

        Parameters
        ----------
        item : `ColumnExpression`
            Item expression to test.

        Returns
        -------
        contains : `ColumnInContainer`
            Boolean column expression that tests for membership in the
            container.
        """
        return ColumnInContainer(item, self)

    @classmethod
    def range_literal(cls, r: range) -> ColumnRangeLiteral:
        """Construct a container expression from a range of indices.

        Parameters
        ----------
        r : `range`
            Range object.

        Returns
        -------
        container : `ColumnRangeLiteral`
            Container expression object representing the range.
        """
        return ColumnRangeLiteral(r)

    @classmethod
    def sequence(
        cls, items: Sequence[ColumnExpression], dtype: type | None = None
    ) -> ColumnExpressionSequence:
        """Construct a container expression from a sequence of item
        expressions.

        Parameters
        ----------
        items : `~collections.abc.Sequence` [ `ColumnExpression` ]
            Sequence of item expressions.
        dtype : `type`, optional
            The Python type of the elements in the container.

        Returns
        -------
        container : `ColumnExpressionSequence`
            Container expression object backed by the given items.
        """
        return ColumnExpressionSequence(items, dtype)


@dataclasses.dataclass(frozen=True)
class ColumnRangeLiteral(ColumnContainer):
    """A container expression backed by a range of integer indices."""

    value: range
    """Range value that backs the expression (`range`).
    """

    dtype: type[int] = int
    """The Python type of the elements in the container (`type` [ `int` ] )."""

    @property
    def columns_required(self) -> Set[ColumnTag]:
        # Docstring inherited.
        return frozenset()

    def __str__(self) -> str:
        return f"[{self.value.start}:{self.value.stop}:{self.value.step}]"

    def is_supported_by(self, engine: Engine) -> bool:
        # Docstring inherited.
        return True


@dataclasses.dataclass(frozen=True)
class ColumnExpressionSequence(ColumnContainer):
    """A container expression backed by a sequence of scalar column
    expressions.
    """

    items: Sequence[ColumnExpression]
    """Sequence of item expressions
    (`~collections.abc.Sequence` [ `ColumnExpression` ]).
    """

    dtype: type | None
    """The Python type of the elements in the container (`type` or `None`)."""

    @property
    @cached_getter
    def columns_required(self) -> Set[ColumnTag]:
        # Docstring inherited.
        result: set[ColumnTag] = set()
        for item in self.items:
            result.update(item.columns_required)
        return result

    def __str__(self) -> str:
        return f"[{', '.join(str(i) for i in self.items)}]"

    def is_supported_by(self, engine: Engine) -> bool:
        # Docstring inherited.
        return all(item.is_supported_by(engine) for item in self.items)


@dataclasses.dataclass(frozen=True)
class ColumnInContainer(Predicate):
    """A boolean column expression that tests whether a scalar column
    expression is present in a container expression.
    """

    item: ColumnExpression
    """Item to be tested (`ColumnExpression`)."""

    container: ColumnContainer
    """Container to be tested (`ColumnContainer`)."""

    @property
    @cached_getter
    def columns_required(self) -> Set[ColumnTag]:
        # Docstring inherited.
        return self.item.columns_required | self.container.columns_required

    def __str__(self) -> str:
        return f"{self.item}∈{self.container}"

    def is_supported_by(self, engine: Engine) -> bool:
        # Docstring inherited.
        return self.item.is_supported_by(engine) and self.container.is_supported_by(engine)

    def as_trivial(self) -> None:
        # Docstring inherited.
        return None



# This file is part of daf_relation.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

import unittest
from collections.abc import Mapping, Set
from typing import Any

from lsst.daf.relation import (
    BinaryOperationRelation,
    ColumnExpression,
    ColumnTag,
    Engine,
    GenericConcreteEngine,
    LeafRelation,
    Materialization,
    Processor,
    Relation,
    SortTerm,
    Transfer,
    UnaryOperationRelation,
    tests,
)


class StringEngine(GenericConcreteEngine[str]):
    """A test Engine whose payloads are just the `str` of their relations."""

    def get_join_identity_payload(self) -> str:
        return "I"

    def get_doomed_payload(self, columns: Set[ColumnTag]) -> str:
        return "0"

    def make_leaf(self, name: str, *columns: ColumnTag, **kwargs: Any) -> Relation:
        return LeafRelation(self, frozenset(columns), name=name, payload=name, **kwargs)


class StringProcessor(Processor):
    """A test subclass of `Processor` that tracks the calls made to its hook
    methods and attaches the `str` of relations as payloads.
    """

    def __init__(self, test_case: ProcessorTestCase):
        self.test_case = test_case
        self.seen: list[str] = []

    def transfer(self, source: Relation, destination: Engine, materialize_as: str | None) -> str:
        self.test_case.check_upstream_payloads(source)
        relation = source.transferred_to(destination)
        if materialize_as is not None:
            relation = relation.materialized(materialize_as)
        result = str(relation)
        self.seen.append(result)
        return result

    def materialize(self, target: Relation, name: str) -> str:
        self.test_case.check_upstream_payloads(target)
        result = str(target.materialized(name))
        self.seen.append(result)
        return result


class ProcessorTestCase(tests.RelationTestCase):
    """Tests for the Processor class."""

    def check_upstream_payloads(
        self,
        relation: Relation,
        materializations_only: bool = False,
        upstream_of_materialization: str | None = None,
        simplifications: Mapping[str, str] | None = None,
    ) -> None:
        """Check that a relation and its upstream tree have the payloads
        that should be attached by `StringProcessor`.

        Parameters
        ----------
        relation : `Relation`
            Relation to check.
        materializations_only : `bool`, optional
            If `True`, only expect leaf and materialization relations to have
            payloads, not transfers, as expected for a tree passed to (but not
            returned by) a `Processor`.
        upstream_of_materialization : `str` | None, optional
            If not `None`, this relation is just upstream of a materialization
            with this name and needs to adjust its expected `str` accordingly
            to include that materialization.
        simplifications : `~collections.abc.Mapping` [ `str`, `str` ]
            Mappings from the original `str` of a relation subtree and the
            simplified form that should have been used to compute the payload
            by `StringProcessor`.
        """
        if simplifications is None:
            simplifications = {}
        if relation.is_join_identity:
            expected_string = "I"
        elif relation.max_rows == 0:
            expected_string = "0"
        else:
            if upstream_of_materialization is not None:
                expected_string = str(relation.materialized(upstream_of_materialization))
            else:
                expected_string = str(relation)
        expected_string = simplifications.get(expected_string, expected_string)
        match relation:
            case LeafRelation():
                self.assertIsNotNone(relation.payload)
                self.assertEqual(relation.payload, expected_string)
            case Materialization():
                self.assertIsNotNone(relation.payload)
                self.assertEqual(relation.payload, expected_string)
                self.check_upstream_payloads(
                    relation.target,
                    materializations_only=materializations_only,
                    upstream_of_materialization=relation.name,
                    simplifications=simplifications,
                )
            case Transfer():
                if materializations_only:
                    self.assertIsNone(relation.payload)
                else:
                    self.assertIsNotNone(relation.payload)
                    self.assertEqual(relation.payload, expected_string)
                self.check_upstream_payloads(
                    relation.target,
                    materializations_only=materializations_only,
                    simplifications=simplifications,
                )
            case UnaryOperationRelation():
                self.check_upstream_payloads(
                    relation.target,
                    materializations_only=materializations_only,
                    simplifications=simplifications,
                )
            case BinaryOperationRelation():
                self.check_upstream_payloads(
                    relation.lhs,
                    materializations_only=materializations_only,
                    simplifications=simplifications,
                )
                self.check_upstream_payloads(
                    relation.rhs,
                    materializations_only=materializations_only,
                    simplifications=simplifications,
                )

    def test_processor(self) -> None:
        """Test the Processor class."""
        # Cook up a three-engine relation tree with pretty some interesting
        # structure to it including some materializations; start with the
        # ingredients.
        engine1 = StringEngine(name="one")
        engine2 = StringEngine(name="two")
        engine3 = StringEngine(name="three")
        a = tests.ColumnTag("a")
        b = tests.ColumnTag("b")
        c = tests.ColumnTag("c")
        d = tests.ColumnTag("d")
        expression = ColumnExpression.reference(b).method("__neg__")
        predicate = ColumnExpression.reference(c).gt(ColumnExpression.literal(0))
        terms = [SortTerm(ColumnExpression.reference(b))]
        leaf1 = engine1.make_leaf("leaf1", a, b)
        leaf2 = engine2.make_leaf("leaf1", a, c)
        leaf3 = engine3.make_leaf("leaf3", a, b, d)
        leaf4 = engine3.make_leaf("leaf4", a, b, d)
        # Build the tree itself while taking snapshops of its str(...)
        # everywhere there's a transfer and/or materialization.
        snapshots = []
        full_tree = (
            leaf2.with_rows_satisfying(predicate).transferred_to(engine1).materialized("materialization1")
        )
        snapshots.append(str(full_tree))
        full_tree = (
            leaf1.with_calculated_column(d, expression)
            .join(full_tree)
            .with_only_columns({a, b, d})
            .transferred_to(engine3)
        )
        snapshots.append(str(full_tree))
        full_tree = full_tree.chain(leaf3).materialized("materialization2")
        snapshots.append(str(full_tree))

        # Chain the full_tree to a what's ultimately a relation with no rows.
        # The Processor will drop these operations without calling its
        # transfer() and materialize() hooks, so they won't appear in its
        # snapshots or the processed_tree it returns.
        trimmed_tree = full_tree
        full_tree = full_tree.chain(
            engine1.make_leaf("doomed_by_join", a, b)
            .join(engine1.make_doomed_relation({b, d}, ["badness"]))
            .transferred_to(engine3)
        )
        # Add some more chains, one of which will simplify before a
        # materialization.
        full_subtree = leaf4.chain(engine3.make_doomed_relation({a, b, d}, ["badness again"])).materialized(
            "materialization3"
        )
        full_tree = full_tree.chain(full_subtree)
        trimmed_tree = trimmed_tree.chain(leaf4)
        # Add a few more operations to both the full_ and trimmed_trees.
        full_tree = full_tree.without_duplicates().sorted(terms)
        trimmed_tree = trimmed_tree.without_duplicates().sorted(terms)
        # Construct and run the Processor, which itself checks some aspects
        # of the algorithm via calls to check_upstream_payloads.
        processor = StringProcessor(self)
        processed_tree = processor.process(full_tree)
        # Check that the snapshots taken by the processor match the ones we
        # took while creating the tree.
        self.assertEqual(processor.seen, snapshots)
        # Check that the processed tree has the same columns, row bounds, and
        # engine as the original.
        self.assertEqual(full_tree.columns, processed_tree.columns)
        self.assertEqual(full_tree.min_rows, processed_tree.min_rows)
        self.assertEqual(full_tree.max_rows, processed_tree.max_rows)
        # Check that the full tree now has payloads for materializations, but
        # not transfers.
        self.check_upstream_payloads(
            full_tree, materializations_only=True, simplifications={str(full_subtree): str(leaf4)}
        )
        # Check that the returned tree how has payloads for materializations
        # and transfers.
        self.check_upstream_payloads(processed_tree, materializations_only=False)
        # Check that the returned tree has the same structure as the trimmed
        # tree.
        self.assert_relations_equal(trimmed_tree, processed_tree)
        # Process the original tree again, which should short-circuit at the
        # last materializations and not call its hooks at all.
        reprocessor = StringProcessor(self)
        reprocessed_tree = reprocessor.process(full_tree)
        self.assertEqual(reprocessor.seen, [])
        self.assertEqual(full_tree.columns, reprocessed_tree.columns)
        self.assertEqual(full_tree.min_rows, reprocessed_tree.min_rows)
        self.assertEqual(full_tree.max_rows, reprocessed_tree.max_rows)


if __name__ == "__main__":
    unittest.main()



# This file is part of daf_relation.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations

__all__ = ("Diagnostics",)

import dataclasses
from collections.abc import Callable

from ._leaf_relation import LeafRelation
from ._marker_relation import MarkerRelation
from ._operation_relations import BinaryOperationRelation, UnaryOperationRelation
from ._operations import Chain, Join, Selection, Slice
from ._relation import Relation


@dataclasses.dataclass
class Diagnostics:
    """A relation-processing algorithm that attempts to explain why a relation
    has no rows.

    The `Diagnostics` class itself is just the type returned by its `run` class
    method, which performs a depth-first tree traversal looking for relations
    that are either known in advance to have no rows (`Relation.max_rows` zero)
    or shown to have no rows via an ``executor`` callable; when present, these
    are then propagated downstream to the root.  Only operations that can
    remove all rows (`UnaryOperation.is_empty_invariant` is `False`) are
    executed when no empty leaf relations are found.x
    """

    is_doomed: bool
    """Whether the given relation will have no rows (`bool`).
    """

    messages: list[str]
    """Messages that explain why the relation will have fewer rows that it
    might have (`list` [ `str` ]).

    This is usually non-empty only when the relation is doomed, but in rare
    cases it can also provide information about what's missing when the
    relation is not doomed, such as when only one of the two branches of a
    `Chain` operation are doomed.
    """

    @classmethod
    def run(cls, relation: Relation, executor: Callable[[Relation], bool] | None = None) -> Diagnostics:
        """Report on whether the given relation has no rows, and if so, why.

        Parameters
        ----------
        relation : `Relation`
            Relation to analyze.
        executor : `Callable`, optional
            If provided, a callable that takes a `Relation` and does some
            engine-specific processing to determine whether it has any rows,
            such as a ``LIMIT 1`` query in SQL.  If not provided, diagnostics
            will be based only on relations with `Relation.max_rows` set to
            zero.

        Returns
        -------
        diagnostics : `Diagnostics`
            Struct containing the diagnostics report.
        """
        match relation:
            case LeafRelation(messages=messages):
                messages = list(messages)
                if relation.max_rows == 0:
                    if not messages:
                        messages.append(f"Relation '{relation!s}' has no rows (static).")
                    return cls(True, messages)
                elif executor is not None and not executor(relation):
                    if not messages:
                        messages.append(f"Relation '{relation!s}' has no rows (executed).")
                    return cls(True, messages)
                else:
                    return cls(False, messages)
            case MarkerRelation(target=target):
                return cls.run(target, executor)
            case UnaryOperationRelation(operation=operation, target=target):
                if (result := cls.run(target, executor)).is_doomed:
                    return result
                if relation.max_rows == 0 and operation.applied_max_rows(target) != 0:
                    result.messages.append(f"Unary operation relation {relation} has no rows (static).")
                    return cls(True, result.messages)
                match operation:
                    case Slice(limit=limit):
                        if limit == 0:
                            result.is_doomed = True
                            result.messages.append(f"Slice with limit=0 applied to '{target!s}'")
                            return result
                    case Selection(predicate=predicate):
                        if predicate.as_trivial() is False:
                            result.is_doomed = True
                            result.messages.append(
                                f"Predicate '{predicate}' is trivially false (applied to '{target!s}')"
                            )
                            return result
                if not operation.is_empty_invariant and executor is not None and not executor(relation):
                    result.is_doomed = True
                    result.messages.append(
                        f"Operation {operation} yields no results when applied to '{target!s}'"
                    )
                    return result
                return result
            case BinaryOperationRelation(operation=operation, lhs=lhs, rhs=rhs):
                lhs_result = cls.run(lhs, executor)
                rhs_result = cls.run(rhs, executor)
                messages = lhs_result.messages + rhs_result.messages
                if relation.max_rows == 0 and operation.applied_max_rows(lhs, rhs) != 0:
                    messages.append(f"Binary operation relation '{relation}' has no rows (static).")
                    return cls(True, messages)
                match operation:
                    case Chain():
                        return cls(lhs_result.is_doomed and rhs_result.is_doomed, messages)
                    case Join(predicate=predicate):
                        if lhs_result.is_doomed or rhs_result.is_doomed:
                            return cls(True, messages)
                        if predicate.as_trivial() is False:
                            messages.append(
                                f"Join predicate '{predicate}' is trivially false in '{relation}'."
                            )
                            return cls(True, messages)
                if executor is not None and not executor(relation):
                    messages.append(f"Operation {operation} yields no results when executed: '{relation!s}'")
                    return cls(True, messages)
                return cls(False, messages)
        raise AssertionError("match should be exhaustive and all branches should return")
